
from contextlib import contextmanager
from ctypes import pythonapi, py_object, c_int
from inspect import stack
from sys import stderr

from .methods import Methods

IVAR_KWARG_NAME = 'ivars'
METHOD_KWARG_NAME = 'methods'
CMETHOD_KWARG_NAMES = ['cmethods','classmethods']
SMETHOD_KWARG_NAMES = ['smethods','staticmethods']
SMEMbER_KWARG_NAMES = ['smembers','staticmembers']
PROP_KWARG_NAMES = ['props','properties']
CACHED_PROP_KWARG_NAMES = ['cached_props','cached_properties']

@contextmanager
def use(o, local_conflict='ignore', **kwargs):
    stk = stack()
    frame_info = stk[2]
    frame = frame_info.frame
    glbls = frame.f_globals
    restore = {}
    if type(o) is dict:
        include = exclude = None
        if kwargs:
            if 'include' in kwargs:
                include = kwargs.pop('include')
                if isinstance(include, str): include = [include]
            if 'exclude' in kwargs:
                exclude = kwargs.pop('exclude')
                if isinstance(exclude, str): exclude = [exclude]
            if kwargs:
                raise ValueError(f'Unexpected kwargs for use(dict): {",".join(kwargs)}')
            if include and exclude:
                raise ValueError('Specify at most one of [include,exclude]')

        for k,v in o.items():
            if include and k not in include: continue
            if exclude and k in exclude: continue
            if k in glbls:
                restore[k] = True, glbls[k]
            else:
                restore[k] = False, None
            if local_conflict != 'ignore':
                if k in frame.f_locals:
                    # TODO: can this be refined to distinguish function vs class vs module scope?
                    # locals can be reliably updated in module scope, afaict.
                    msg = f'var {k} found in local scope; setting global may not take effect'
                    if local_conflict == 'warn':
                        stderr.write('%s\n' % msg)
                    elif local_conflict in ['raise','err','error']:
                        raise RuntimeError(msg)
                    else:
                        raise ValueError(f'Unrecognized `local_conflict` value: {local_conflict}')
            glbls[k] = v
    else:
        # Parse kwargs
        if 'include' in kwargs:
            ivars = False
            methods = False
            cmethods = False
            smethods = False
            smembers = False
            properties = False
            cached_properties = False
            include = kwargs['include']
            if isinstance(include, str): include = [include]
            for k in include:
                if k == IVAR_KWARG_NAME: ivars = True
                if k == METHOD_KWARG_NAME: methods = True
                if k in CMETHOD_KWARG_NAMES: cmethods = True
                if k in SMETHOD_KWARG_NAMES: smethods = True
                if k in SMEMbER_KWARG_NAMES: smembers = True
                if k in PROP_KWARG_NAMES: properties = True
                if k in CACHED_PROP_KWARG_NAMES: cached_properties = True
        else:
            ivars = True
            methods = True
            cmethods = True
            smethods = True
            smembers = True
            properties = True
            cached_properties = True
            if 'exclude' in kwargs:
                exclude = kwargs['exclude']
                if isinstance(exclude, str): exclude = [exclude]
                for k in exclude:
                    if k == IVAR_KWARG_NAME: ivars = False
                    if k == METHOD_KWARG_NAME: methods = False
                    if k in CMETHOD_KWARG_NAMES: cmethods = False
                    if k in SMETHOD_KWARG_NAMES: smethods = False
                    if k in SMEMbER_KWARG_NAMES: smembers = False
                    if k in PROP_KWARG_NAMES: properties = False
                    if k in CACHED_PROP_KWARG_NAMES: cached_properties = False
            else:
                for k,v in kwargs.items():
                    if k == IVAR_KWARG_NAME: ivars = bool(v)
                    if k == METHOD_KWARG_NAME: methods = bool(v)
                    if k in CMETHOD_KWARG_NAMES: cmethods = bool(v)
                    if k in SMETHOD_KWARG_NAMES: smethods = bool(v)
                    if k in SMEMbER_KWARG_NAMES: smembers = bool(v)
                    if k in PROP_KWARG_NAMES:
                        if v == 'all':
                            pass
                        elif v == 'cached':
                            properties = False
                        elif v in ['none','no']:
                            properties = False
                            cached_properties = False
                        else:
                            raise ValueError(f'Unexpcted "properties" value: {properties}; choices: [all,cached,none/no]')

        m = Methods(o)

        assign = set()
        if             ivars: assign.update(m.            ivars)
        if           methods: assign.update(m.          methods)
        if          cmethods: assign.update(m.         cmethods)
        if          smethods: assign.update(m.         smethods)
        if          smembers: assign.update(m.         smembers)
        if cached_properties: assign.update(m.cached_properties)
        if        properties: assign.update(m.       properties)

        for k in assign:
            if k in glbls:
                restore[k] = True, glbls[k]
            else:
                restore[k] = False, None
            glbls[k] = getattr(o, k)

    pythonapi.PyFrame_LocalsToFast(py_object(frame), c_int(1))
    del frame
    yield

    if restore:
        frame = frame_info.frame
        glbls = frame.f_globals
        for k, (existed, v) in restore.items():
            if existed:
                glbls[k] = v
            else:
                del glbls[k]
        pythonapi.PyFrame_LocalsToFast(py_object(frame), c_int(1))
        del frame





#!/usr/bin/env python
# coding: utf-8

# In[ ]:


from os import chdir, getcwd
from pathlib import Path


class cd:
    """Context manager for changing the current working directory"""
    def __init__(self, path):
        path = Path(path)
        self.path = path.expanduser()

    def __enter__(self):
        self.prevPath = Path(getcwd())
        chdir(str(self.path))

    def __exit__(self, etype, value, traceback):
        chdir(str(self.prevPath))





from os.path import join
from os import getcwd, remove
from tempfile import NamedTemporaryFile

from .process import run, sh

class File:
    def __init__(self, path=None, tag=None, rm=None, copy_dir=None, **kwargs):
        if kwargs:
            if 'dir' in kwargs:
                dir = kwargs.pop('dir')
            else:
                dir = getcwd()
            if kwargs:
                raise ValueError(f'Unrecognized kwargs: {kwargs}')
        else:
            dir = getcwd()
        if path:
            self.rm = bool(rm)
        else:
            path = NamedTemporaryFile(prefix='Dockerfile.', dir=dir).name
            self.rm = rm != False
        self.path = path
        self.fd = None
        self.tag = tag
        self.dir = dir
        self.copy_dir = copy_dir

    def write(self, *lines, open_ok=True):
        if not self.fd:
            if not open_ok:
                raise RuntimeError(f"Refusing to write to {self.path} that's not already open")
            self.fd = open(self.path,'w')
        for line in lines:
            self.fd.write('%s\n' % line)

    def close(self, closed_ok=False):
        if self.fd:
            self.fd.close()
            self.fd = None
        elif not closed_ok:
            raise RuntimeError(f"Can't close {self.path}; not open")

    def build(self, tag=None, dir=None, **build_args):
        self.close()

        tag = tag or self.tag
        dir = dir or self.dir or '.'
        if not tag:
            raise ValueError(f'Missing tag for image build')

        run(
            'docker','build',
            '-f',self.path,
            '-t',tag,
            [ ['--build-arg',f'{k}={v}'] for k,v in build_args.items() ],
            dir
        )
        if self.rm:
            remove(self.path)

        return Image(tag, self)

    def COPY(self, *args, **kwargs):
        (*srcs, dst) = args
        if 'dir' in kwargs:
            dir = kwargs.pop('dir')
            if kwargs:
                raise ValueError(f'Unexpected kwargs: {",".join(kwargs)}')
        else:
            dir = self.copy_dir
        if dir:
            srcs = [ join(dir, src) for src in srcs ]
        self.write(f'COPY {" ".join([*srcs, dst])}')

    def ENV(self, *args, **kwargs):
        for arg in args:
            if isinstance(arg, str):
                self.write(f'ENV {arg}')
            elif isinstance(arg, dict):
                self.ENV(**arg)
            else:
                raise ValueError(f'Unrecognized argument to ENV(): {arg}')
        for k,v in kwargs.items():
            self.write(f'ENV {k}={v}')

    def NOTE(self, *lines):
        self.write(*[f'# {line}' for line in lines])

    def RUN(self, *cmds):
        self.write('RUN %s' % ' \\\n && '.join(cmds))

    def ARG(self, k, v=None):
        if v is None:
            self.write(f'ARG {k}')
        else:
            self.write(f'ARG {k}={v}')

    def FROM(self, repo, tag=None, registry=None):
        if registry:
            repo = f'{registry}/{repo}'
        if tag:
            repo = f'{repo}:{tag}'
        self.write(f'FROM {repo}')

    def LN(self): self.write('')

    def WORKDIR(self, dst='/'): self.write(f'WORKDIR {dst}')

    def ENTRYPOINT(self, arg): self.write(f'ENTRYPOINT {arg}')

    def USER(self, user, group=None):
        if group:
            self.write(f'USER {user}:{group}')
        else:
            self.write(f'USER {user}')

class Image:
    def __init__(self, url, file=None):
        self.url = url
        self.file = file

    def run(self, name=None, rm=False,):
        sh(
            'docker','run',
            ['--name',name] if name else None,
            '--rm' if rm else None,
            self.url,
        )
